@@ -7,6 +7,7 @@ module Test.Hls.Util
   (  -- * Test Capabilities
       codeActionResolveCaps
     , codeActionNoResolveCaps
+    , codeActionNoInlayHintsCaps
     , codeActionSupportCaps
     , expectCodeAction
     -- * Environment specifications
@@ -35,6 +36,7 @@ module Test.Hls.Util
     , inspectCodeAction
     , inspectCommand
     , inspectDiagnostic
+    , inspectDiagnosticAny
     , waitForDiagnosticsFrom
     , waitForDiagnosticsFromSource
     , waitForDiagnosticsFromSourceWithTimeout
@@ -45,7 +47,7 @@ module Test.Hls.Util
     -- * Extract positions from input file.
     , extractCursorPositions
     , mkParameterisedLabel
-    , trimming
+    , __i
   )
 where
 
@@ -79,11 +81,11 @@ import           Test.Tasty.ExpectedFailure               (expectFailBecause,
 import           Test.Tasty.HUnit                         (assertFailure)
 
 import qualified Data.List                                as List
+import           Data.String.Interpolate                  (__i)
 import qualified Data.Text.Internal.Search                as T
 import qualified Data.Text.Utf16.Rope.Mixed               as Rope
 import           Development.IDE.Plugin.Completions.Logic (getCompletionPrefixFromRope)
 import           Development.IDE.Plugin.Completions.Types (PosPrefixInfo (..))
-import           NeatInterpolation                        (trimming)
 
 noLiteralCaps :: ClientCapabilities
 noLiteralCaps = def & L.textDocument ?~ textDocumentCaps
@@ -107,6 +109,12 @@ codeActionNoResolveCaps :: ClientCapabilities
 codeActionNoResolveCaps = Test.fullLatestClientCaps
                           & (L.textDocument . _Just . L.codeAction . _Just . L.resolveSupport) .~ Nothing
                           & (L.textDocument . _Just . L.codeAction . _Just . L.dataSupport . _Just) .~ False
+
+codeActionNoInlayHintsCaps :: ClientCapabilities
+codeActionNoInlayHintsCaps = Test.fullLatestClientCaps
+                          & (L.textDocument . _Just . L.codeAction . _Just . L.resolveSupport) .~ Nothing
+                          & (L.textDocument . _Just . L.codeAction . _Just . L.dataSupport . _Just) .~ False
+                          & (L.textDocument . _Just . L.inlayHint) .~ Nothing
 -- ---------------------------------------------------------------------
 -- Environment specification for ignoring tests
 -- ---------------------------------------------------------------------
@@ -240,6 +248,10 @@ inspectDiagnostic :: [Diagnostic] -> [T.Text] -> IO Diagnostic
 inspectDiagnostic diags s = onMatch diags (\ca -> all (`T.isInfixOf` (ca ^. L.message)) s) err
     where err = "expected diagnostic matching '" ++ show s ++ "' but did not find one"
 
+inspectDiagnosticAny :: [Diagnostic] -> [T.Text] -> IO Diagnostic
+inspectDiagnosticAny diags s = onMatch diags (\ca -> any (`T.isInfixOf` (ca ^. L.message)) s) err
+    where err = "expected diagnostic matching one of'" ++ show s ++ "' but did not find one"
+
 expectDiagnostic :: [Diagnostic] -> [T.Text] -> IO ()
 expectDiagnostic diags s = void $ inspectDiagnostic diags s
 
